/*
 * Copyright 2005, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * Contributors: Georg Wittenburg
 */

#include "main.h"
#include "eeprom_gen.h"
#include "y.tab.h"

/**
 * This file generates the EEPROM bytecode.
 **/

// Two macros to allow to read/write access to an UINT16 in memory, i.e. two
// adjacent bytes, in one operation.
#define MEM_TO_UINT16(pointer, offset, uint16_ptr) \
	*uint16_ptr = (UINT16) ((UINT8) pointer[offset]); \
	*uint16_ptr |= ((UINT16) ((UINT8) pointer[offset + 1])) << 8;
#define UINT16_TO_MEM(uint16, pointer, offset) \
	pointer[offset] = (UINT8) ((UINT16) uint16); \
	pointer[offset + 1] = (UINT8) (((UINT16) uint16) >> 8);

typedef struct {
	void* buffer;
	int counter;
} accumulator_t;

// The memory region representing the EEPROM image.
char* eeprom_image;
// The rulset currently being worked on.
rc_ruleset_t* cur_ruleset;

// Offsets and counters of items written to the EEPROM image.
int rules_first, rules_last, rules_used;
int facts_first, facts_last, facts_used;
int properties_first, properties_last, properties_used;
int conditions_first, conditions_last, conditions_used;
int statements_first, statements_last, statements_used;
int slots_first, slots_last, slots_used;
int expressions_first, expressions_last, expressions_used;
int initializers_first, initializers_last, initializers_used;
int const_variables_first, const_variables_last, const_variables_used;

/*
 * Identifiers are strings in the source code, but in order to save memory,
 * they are hashed and only a unique ID is written to the EEPROM image.
 */
GHashTable* identifiers;
// The unique ID is generated simply by counting how many identifiers have been
// hashed and using that value.
UINT16 identifier_count;

// Checks a memory region whether it contains a particular search_item and
// returns the address if it does.
void* memfind(void* search_item, void* items_first, void* items_last, size_t item_size, int item_count) {
	void* addr;
	for(addr = items_first; addr < items_last; addr += item_size) {
		if((memcmp(search_item, addr, (item_size * item_count)) == 0) && (addr != search_item))
			return addr;
	}
	return NULL;
}

UINT16 eeprom_gen_identifier(char* identifier) {
	if(strcmp(identifier, rc_name_this.identifier) == 0)
		return RLS_SLOT_NAME_THIS;
	gpointer value = g_hash_table_lookup(identifiers, identifier);
	if(value == NULL) {
		g_hash_table_insert(identifiers, identifier, (gpointer) ((int) ++identifier_count));
		if(debug)
			fprintf(stderr, "Adding new identifier bytecode value 0x%x for \"%s\".\n", identifier_count, identifier);
		return identifier_count;
	} else {
		return ((UINT16) ((int) value));
	}
}

UINT16 eeprom_gen_name(rc_name_t* rc_name) {
	if(rc_name == &rc_name_this)
		return RLS_SLOT_NAME_THIS;
	else
		return eeprom_gen_identifier(rc_name->string);
}

UINT16 eeprom_gen_slot(rc_slot_t* rc_slot);

void eeprom_gen_variable_at(rc_variable_t* rc_variable, variable_t* variable) {
	switch(rc_variable->type) {
		case RC_BOOL:
			variable->type = RLS_VARIABLE_TYPE_BOOL;
			if(rc_variable->value)
				variable->value = ((UINT16) True);
			else
				variable->value = ((UINT16) False);
			break;
		case RC_INTEGER:
			variable->type = RLS_VARIABLE_TYPE_INTEGER;
			variable->value = ((UINT16) ((int) rc_variable->value));
			break;
		case RC_STRING:
			variable->type = RLS_VARIABLE_TYPE_STRING;
			{
				char* str = (char*) rc_variable->value;
				UINT16 i;
				MEM_TO_UINT16(str, 0, &i);
				if(strlen(str) == 0) {
					variable->value = 0x0000;
				} else if(strlen(str) <= 2) {
					// strlen(1) is no problem as the str is \0 terminated.
					variable->value = i;
				} else {
					fprintf(stderr, "Warning: Truncating string \"%s\" to first two chars for EEPROM.\n", str);
					variable->value = i;
				}
			}
			break;
		case RC_SLOT:
			variable->type = RLS_VARIABLE_TYPE_SLOT;
			variable->value = eeprom_gen_slot((rc_slot_t*) rc_variable->value);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid variable type %d.\n", rc_variable->type);
			exit(RC_EINTERNAL);
	}
}

UINT16 eeprom_gen_const_variable(rc_variable_t* rc_variable) {
	UINT16 eeprom_address = const_variables_first + (const_variables_used++ * sizeof(variable_t));
	variable_t* variable = (variable_t*) malloc_zero(sizeof(variable_t));

	eeprom_gen_variable_at(rc_variable, variable);

	if(optimize_bytecode) {
		// Optimize bytecode by reusing previously written items that
		// are exact duplicates of the current item.
		void* addr = memfind(variable, eeprom_image + const_variables_first, eeprom_image + const_variables_first + ((const_variables_used - 1) * sizeof(variable_t)), sizeof(variable_t), 1);
		if(addr) {
			free(variable);
			const_variables_used--;
			return (UINT16) (addr - ((void*) eeprom_image));
		}
	}

	memcpy(eeprom_image + eeprom_address, variable, sizeof(variable_t));
	free(variable);
	return eeprom_address;
}

UINT16 eeprom_gen_condition(rc_condition_t* rc_condition, int in_list_context);
void eeprom_foreach_gen_condition(gpointer data, gpointer user_data);

UINT16 eeprom_gen_slot(rc_slot_t* rc_slot) {
	UINT16 eeprom_address = slots_first + (slots_used++ * sizeof(slot_t));
	slot_t* slot = (slot_t*) malloc_zero(sizeof(slot_t));
	int slot_conditions_count = g_list_length(rc_slot->conditions);

	slot->factName = eeprom_gen_name(rc_slot->name);
	if(rc_slot->key != NULL)
		slot->propertyKey = eeprom_gen_identifier(rc_slot->key);
	else
		slot->propertyKey = RLS_FIELD_UNUSED;
	if(slot_conditions_count) {
		slot->conditions_first = conditions_first + (conditions_used * sizeof(condition_t));
		slot->conditions_last = slot->conditions_first + ((slot_conditions_count - 1) * sizeof(condition_t));
	} else {
		slot->conditions_first = RLS_FIELD_UNUSED;
		slot->conditions_last = RLS_FIELD_UNUSED;
	}

	if(optimize_bytecode) {
		if(slot_conditions_count) {
			// Optimize bytecode by reusing a previously written
			// list of items that matches the list created by the
			// current item as one of its subitems.

			if(debug)
				fprintf(stderr, "Slot {\"%s\" \"%s\"} at 0x%x with %d condition(s):\n", rc_slot->name->string, rc_slot->key, eeprom_address, slot_conditions_count);

			accumulator_t* accumulator = malloc_zero(sizeof(accumulator_t));
			accumulator->buffer = malloc_zero(slot_conditions_count * sizeof(condition_t));
			accumulator->counter = 0;
			g_list_foreach(rc_slot->conditions, eeprom_foreach_gen_condition, accumulator);
			void* addr = memfind(accumulator->buffer, eeprom_image + conditions_first, eeprom_image + conditions_first + (conditions_used * sizeof(condition_t)), sizeof(condition_t), slot_conditions_count);
			if(addr) {
				slot->conditions_first = (UINT16) (addr - ((void*) eeprom_image));
				slot->conditions_last = slot->conditions_first + ((slot_conditions_count - 1) * sizeof(condition_t));
				if(debug)
					fprintf(stderr, "\tReusing condition(s) at 0x%x, slot points from 0x%x to 0x%x.\n", (int) (addr - ((void*) eeprom_image)), slot->conditions_first, slot->conditions_last);
			} else {
				memcpy(eeprom_image + conditions_first + (conditions_used * sizeof(condition_t)), accumulator->buffer, slot_conditions_count * sizeof(condition_t));
				// We need to recalculate the location of the
				// conditions, because we might have generated
				// more slots recursively when generating the
				// conditions, thus rendering the old data
				// incorrect.
				slot->conditions_first = conditions_first + (conditions_used * sizeof(condition_t));
				slot->conditions_last = slot->conditions_first + ((slot_conditions_count - 1) * sizeof(condition_t));
				if(debug)
					fprintf(stderr, "\tGenerating new condition(s) at 0x%x, slot points from 0x%x to 0x%x.\n", conditions_first + (conditions_used * sizeof(condition_t)), slot->conditions_first, slot->conditions_last);
				conditions_used += slot_conditions_count;
			}
			free(accumulator->buffer);
			free(accumulator);
		}

		// Optimize bytecode by reusing previously written items that
		// are exact duplicates of the current item.
		void* addr = memfind(slot, eeprom_image + slots_first, eeprom_image + slots_first + ((slots_used - 1) * sizeof(slot_t)), sizeof(slot_t), 1);
		if(addr) {
			free(slot);
			slots_used--;
			return (UINT16) (addr - ((void*) eeprom_image));
		}
	} else {
		g_list_foreach(rc_slot->conditions, eeprom_foreach_gen_condition, NULL);
	}

	memcpy(eeprom_image + eeprom_address, slot, sizeof(slot_t));
	free(slot);
	return eeprom_address;
}

UINT16 eeprom_gen_unary_operation(int unary_operation) {
	switch(unary_operation) {
		case RC_COUNT:
			return RLS_EXPRESSION_UNARY_OP_COUNT;
		case RC_SUM:
			return RLS_EXPRESSION_UNARY_OP_SUM;
		case RC_PRODUCT:
			return RLS_EXPRESSION_UNARY_OP_PRODUCT;
		case RC_MINIMUM:
			return RLS_EXPRESSION_UNARY_OP_MINIMUM;
		case RC_MAXIMUM:
			return RLS_EXPRESSION_UNARY_OP_MAXIMUM;
		case RC_SQUARE_ROOT:
			return RLS_EXPRESSION_UNARY_OP_SQUARE_ROOT;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid unary arithmetic operation %d.\n", unary_operation);
			exit(RC_EINTERNAL);
	}
}

UINT16 eeprom_gen_binary_operation(int binary_operation) {
	switch(binary_operation) {
		case RC_ADD:
			return RLS_EXPRESSION_BINARY_OP_ADD;
		case RC_SUBTRACT:
			return RLS_EXPRESSION_BINARY_OP_SUBTRACT;
		case RC_MULTIPLY:
			return RLS_EXPRESSION_BINARY_OP_MULTIPLY;
		case RC_DIVIDE:
			return RLS_EXPRESSION_BINARY_OP_DIVIDE;
		case RC_TO_THE_POWER_OF:
			return RLS_EXPRESSION_BINARY_OP_TO_THE_POWER_OF;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid binary arithmetic operation %d.\n", binary_operation);
			exit(RC_EINTERNAL);
	}
}

UINT16 eeprom_gen_expression(rc_expression_t* rc_expression, int in_list_context);

void eeprom_gen_expression_at(rc_expression_t* rc_expression, expression_t* expression) {
	switch(rc_expression->type) {
		case RC_CONSTANT:
			expression->type_AND_operation = RLS_EXPRESSION_TYPE_CONSTANT;
			expression->variable_OR_left_expression = eeprom_gen_const_variable(rc_expression->constant_variable);
			expression->UNUSED_OR_right_expression = RLS_FIELD_UNUSED;
			break;
		case RC_UNARY_OPERATION:
			expression->type_AND_operation = (RLS_EXPRESSION_TYPE_UNARY | eeprom_gen_unary_operation(rc_expression->unary_operation));
			expression->variable_OR_left_expression = eeprom_gen_expression(rc_expression->unary_expression, 0);
			expression->UNUSED_OR_right_expression = RLS_FIELD_UNUSED;
			break;
		case RC_BINARY_OPERATION:
			expression->type_AND_operation = (RLS_EXPRESSION_TYPE_BINARY | eeprom_gen_binary_operation(rc_expression->binary_operation));
			expression->variable_OR_left_expression = eeprom_gen_expression(rc_expression->left_binary_expression, 0);
			expression->UNUSED_OR_right_expression = eeprom_gen_expression(rc_expression->right_binary_expression, 0);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid expression type %d.\n", rc_expression->type);
			exit(RC_EINTERNAL);
	}
}

UINT16 eeprom_gen_expression(rc_expression_t* rc_expression, int in_list_context) {
	UINT16 eeprom_address = expressions_first + (expressions_used++ * sizeof(expression_t));
	expression_t* expression = (expression_t*) malloc_zero(sizeof(expression_t));

	eeprom_gen_expression_at(rc_expression, expression);

	if(optimize_bytecode && !in_list_context) {
		// Optimize bytecode by reusing previously written items that
		// are exact duplicates of the current item.
		void* addr = memfind(expression, eeprom_image + expressions_first, eeprom_image + expressions_first + ((expressions_used - 1) * sizeof(expression_t)), sizeof(expression_t), 1);
		if(addr) {
			free(expression);
			expressions_used--;
			return (UINT16) (addr - ((void*) eeprom_image));
		}
	}

	memcpy(eeprom_image + eeprom_address, expression, sizeof(expression_t));
	free(expression);
	return eeprom_address;
}

void eeprom_foreach_gen_expression(gpointer data, gpointer user_data) {
	if(user_data == NULL) {
		eeprom_gen_expression((rc_expression_t*) data, 1);
	} else {
		accumulator_t* accumulator = (accumulator_t*) user_data;
		eeprom_gen_expression_at((rc_expression_t*) data, accumulator->buffer + (accumulator->counter++ * sizeof(expression_t)));
	}
}

UINT16 eeprom_gen_comparison_operation(int comparison_operation) {
	switch(comparison_operation) {
		case RC_EQUALS:
			return RLS_CONDITION_OP_EQUALS;
		case RC_NOT_EQUALS:
			return RLS_CONDITION_OP_NOT_EQUALS;
		case RC_LESS_THAN:
			return RLS_CONDITION_OP_LESS_THAN;
		case RC_GREATER_THAN:
			return RLS_CONDITION_OP_GREATER_THAN;
		case RC_LESS_OR_EQUAL:
			return RLS_CONDITION_OP_LESS_OR_EQUAL;
		case RC_GREATER_OR_EQUAL:
			return RLS_CONDITION_OP_GREATER_OR_EQUAL;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid comparison operation type %d.\n", comparison_operation);
			exit(RC_EINTERNAL);
	}
}

void eeprom_gen_condition_at(rc_condition_t* rc_condition, condition_t* condition) {
	switch(rc_condition->type) {
		case RC_EXISTS:
			condition->type_AND_condition_operation = RLS_CONDITION_TYPE_EXISTS;
			condition->slot_OR_left_expression = eeprom_gen_slot(rc_condition->exists_slot);
			condition->UNUSED_OR_right_expression = RLS_FIELD_UNUSED;
			break;
		case RC_EVAL:
			condition->type_AND_condition_operation = (RLS_CONDITION_TYPE_EVAL | eeprom_gen_comparison_operation(rc_condition->eval_comparison_operation));
			condition->slot_OR_left_expression = eeprom_gen_expression(rc_condition->eval_left_expression, 0);
			condition->UNUSED_OR_right_expression = eeprom_gen_expression(rc_condition->eval_right_expression, 0);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid condition type %d.\n", rc_condition->type);
			exit(RC_EINTERNAL);
	}
}

UINT16 eeprom_gen_condition(rc_condition_t* rc_condition, int in_list_context) {
	UINT16 eeprom_address = conditions_first + (conditions_used++ * sizeof(condition_t));
	condition_t* condition = (condition_t*) malloc_zero(sizeof(condition_t));

	eeprom_gen_condition_at(rc_condition, condition);

	if(optimize_bytecode && !in_list_context) {
		// Optimize bytecode by reusing previously written items that
		// are exact duplicates of the current item.
		void* addr = memfind(condition, eeprom_image + conditions_first, eeprom_image + conditions_first + ((conditions_used - 1) * sizeof(condition_t)), sizeof(condition_t), 1);
		if(addr) {
			free(condition);
			conditions_used--;
			return (UINT16) (addr - ((void*) eeprom_image));
		}
	}

	memcpy(eeprom_image + eeprom_address, condition, sizeof(condition_t));
	free(condition);
	return eeprom_address;
}

void eeprom_foreach_gen_condition(gpointer data, gpointer user_data) {
	if(user_data == NULL) {
		eeprom_gen_condition((rc_condition_t*) data, 1);
	} else {
		accumulator_t* accumulator = (accumulator_t*) user_data;
		eeprom_gen_condition_at((rc_condition_t*) data, accumulator->buffer + (accumulator->counter++ * sizeof(condition_t)));
	}
}

void eeprom_gen_initializer_at(rc_initializer_t* rc_initializer, initializer_t* initializer) {
	initializer->key = eeprom_gen_identifier(rc_initializer->key);
	initializer->expression = eeprom_gen_expression(rc_initializer->expression, 0);
}

UINT16 eeprom_gen_initializer(rc_initializer_t* rc_initializer, int in_list_context) {
	UINT16 eeprom_address = initializers_first + (initializers_used++ * sizeof(initializer_t));
	initializer_t* initializer = (initializer_t*) malloc_zero(sizeof(initializer_t));

	eeprom_gen_initializer_at(rc_initializer, initializer);

	if(optimize_bytecode && !in_list_context) {
		// Optimize bytecode by reusing previously written items that
		// are exact duplicates of the current item.
		void* addr = memfind(initializer, eeprom_image + initializers_first, eeprom_image + initializers_first + ((initializers_used - 1) * sizeof(initializer_t)), sizeof(initializer_t), 1);
		if(addr) {
			free(initializer);
			initializers_used--;
			return (UINT16) (addr - ((void*) eeprom_image));
		}
	}

	memcpy(eeprom_image + eeprom_address, initializer, sizeof(initializer_t));
	free(initializer);
	return eeprom_address;
}

void eeprom_foreach_gen_initializer(gpointer data, gpointer user_data) {
	if(user_data == NULL) {
		eeprom_gen_initializer((rc_initializer_t*) data, 1);
	} else {
		accumulator_t* accumulator = (accumulator_t*) user_data;
		eeprom_gen_initializer_at((rc_initializer_t*) data, accumulator->buffer + (accumulator->counter++ * sizeof(initializer_t)));
	}
}

void eeprom_gen_statement_at(rc_statement_t* rc_statement, statement_t* statement) {
	switch(rc_statement->type) {
		case RC_DEFINE:
			{
				int define_initializers_count = g_list_length(rc_statement->define_initializers);
				statement->type = RLS_STATEMENT_TYPE_DEFINE;
				statement->name_OR_slot_OR_identifier = eeprom_gen_name(rc_statement->define_name);
				if(define_initializers_count) {
					statement->parameters_first_OR_left_expression_OR_UNUSED = initializers_first + (initializers_used * sizeof(initializer_t));
					statement->parameters_last_OR_right_expression_OR_UNUSED = statement->parameters_first_OR_left_expression_OR_UNUSED + ((define_initializers_count - 1) * sizeof(initializer_t));
				} else {
					statement->parameters_first_OR_left_expression_OR_UNUSED = RLS_FIELD_UNUSED;
					statement->parameters_last_OR_right_expression_OR_UNUSED = RLS_FIELD_UNUSED;
				}
				if(optimize_bytecode && define_initializers_count) {
					// Optimize bytecode by reusing a previously written
					// list of items that matches the list created by the
					// current item as one of its subitems.
					accumulator_t* accumulator = malloc_zero(sizeof(accumulator_t));
					accumulator->buffer = malloc_zero(define_initializers_count * sizeof(initializer_t));
					accumulator->counter = 0;
					g_list_foreach(rc_statement->define_initializers, eeprom_foreach_gen_initializer, accumulator);
					void* addr = memfind(accumulator->buffer, eeprom_image + initializers_first, eeprom_image + initializers_first + (initializers_used * sizeof(initializer_t)), sizeof(initializer_t), define_initializers_count);
					if(addr) {
						statement->parameters_first_OR_left_expression_OR_UNUSED = (UINT16) (addr - ((void*) eeprom_image));
						statement->parameters_last_OR_right_expression_OR_UNUSED = statement->parameters_first_OR_left_expression_OR_UNUSED + ((define_initializers_count - 1) * sizeof(initializer_t));
					} else {
						memcpy(eeprom_image + initializers_first + (initializers_used * sizeof(initializer_t)), accumulator->buffer, define_initializers_count * sizeof(initializer_t));
						initializers_used += define_initializers_count;
					}
					free(accumulator->buffer);
					free(accumulator);
				} else {
					g_list_foreach(rc_statement->define_initializers, eeprom_foreach_gen_initializer, NULL);
				}
			}
			break;
		case RC_RETRACT:
			statement->type = RLS_STATEMENT_TYPE_RETRACT;
			statement->name_OR_slot_OR_identifier = eeprom_gen_slot(rc_statement->retract_slot);
			statement->parameters_first_OR_left_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			statement->parameters_last_OR_right_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			break;
		case RC_COPY:
			statement->type = RLS_STATEMENT_TYPE_COPY;
			statement->name_OR_slot_OR_identifier = eeprom_gen_slot(rc_statement->copy_slot);
			statement->parameters_first_OR_left_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			statement->parameters_last_OR_right_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			break;
		case RC_SEND:
			statement->type = RLS_STATEMENT_TYPE_SEND;
			statement->name_OR_slot_OR_identifier = eeprom_gen_slot(rc_statement->send_slot);
			statement->parameters_first_OR_left_expression_OR_UNUSED = eeprom_gen_expression(rc_statement->send_address_expression, 0);
			statement->parameters_last_OR_right_expression_OR_UNUSED = eeprom_gen_expression(rc_statement->send_power_expression, 0);
			break;
		case RC_SET:
			statement->type = RLS_STATEMENT_TYPE_SET;
			statement->name_OR_slot_OR_identifier = eeprom_gen_slot(rc_statement->set_slot);
			statement->parameters_first_OR_left_expression_OR_UNUSED = eeprom_gen_expression(rc_statement->set_expression, 0);
			statement->parameters_last_OR_right_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			break;
		case RC_FLUSH:
			statement->type = RLS_STATEMENT_TYPE_FLUSH;
			statement->name_OR_slot_OR_identifier = eeprom_gen_slot(rc_statement->flush_slot);
			statement->parameters_first_OR_left_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			statement->parameters_last_OR_right_expression_OR_UNUSED = RLS_FIELD_UNUSED;
			break;
		case RC_CALL:
			{
				int call_expressions_count = g_list_length(rc_statement->call_expressions);
				statement->type = RLS_STATEMENT_TYPE_CALL;
				statement->name_OR_slot_OR_identifier = eeprom_gen_identifier(rc_statement->call_identifier);
				if(call_expressions_count) {
					statement->parameters_first_OR_left_expression_OR_UNUSED = expressions_first + (expressions_used * sizeof(expression_t));
					statement->parameters_last_OR_right_expression_OR_UNUSED = statement->parameters_first_OR_left_expression_OR_UNUSED + ((call_expressions_count - 1) * sizeof(expression_t));
				} else {
					statement->parameters_first_OR_left_expression_OR_UNUSED = RLS_FIELD_UNUSED;
					statement->parameters_last_OR_right_expression_OR_UNUSED = RLS_FIELD_UNUSED;
				}
				if(optimize_bytecode && call_expressions_count) {
					// Optimize bytecode by reusing a previously written
					// list of items that matches the list created by the
					// current item as one of its subitems.
					accumulator_t* accumulator = malloc_zero(sizeof(accumulator_t));
					accumulator->buffer = malloc_zero(call_expressions_count * sizeof(expression_t));
					accumulator->counter = 0;
					g_list_foreach(rc_statement->call_expressions, eeprom_foreach_gen_expression, accumulator);
					void* addr = memfind(accumulator->buffer, eeprom_image + expressions_first, eeprom_image + expressions_first + (expressions_used * sizeof(expression_t)), sizeof(expression_t), call_expressions_count);
					if(addr) {
						statement->parameters_first_OR_left_expression_OR_UNUSED = (UINT16) (addr - ((void*) eeprom_image));
						statement->parameters_last_OR_right_expression_OR_UNUSED = statement->parameters_first_OR_left_expression_OR_UNUSED + ((call_expressions_count - 1) * sizeof(expression_t));
					} else {
						memcpy(eeprom_image + expressions_first + (expressions_used * sizeof(expression_t)), accumulator->buffer, call_expressions_count * sizeof(expression_t));
						expressions_used += call_expressions_count;
					}
					free(accumulator->buffer);
					free(accumulator);
				} else {
					g_list_foreach(rc_statement->call_expressions, eeprom_foreach_gen_expression, NULL);
				}
			}
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid statement type %d.\n", rc_statement->type);
			exit(RC_EINTERNAL);
	}
}

UINT16 eeprom_gen_statement(rc_statement_t* rc_statement, int in_list_context) {
	UINT16 eeprom_address = statements_first + (statements_used++ * sizeof(statement_t));
	statement_t* statement = (statement_t*) malloc_zero(sizeof(statement_t));

	eeprom_gen_statement_at(rc_statement, statement);

	if(optimize_bytecode && !in_list_context) {
		// Optimize bytecode by reusing previously written items that
		// are exact duplicates of the current item.
		void* addr = memfind(statement, eeprom_image + statements_first, eeprom_image + statements_first + ((statements_used - 1) * sizeof(statement_t)), sizeof(statement_t), 1);
		if(addr) {
			free(statement);
			statements_used--;
			return (UINT16) (addr - ((void*) eeprom_image));
		}
	}

	memcpy(eeprom_image + eeprom_address, statement, sizeof(statement_t));
	free(statement);
	return eeprom_address;
}

void eeprom_foreach_gen_statement(gpointer data, gpointer user_data) {
	if(user_data == NULL) {
		eeprom_gen_statement((rc_statement_t*) data, 1);
	} else {
		accumulator_t* accumulator = (accumulator_t*) user_data;
		eeprom_gen_statement_at((rc_statement_t*) data, accumulator->buffer + (accumulator->counter++ * sizeof(statement_t)));
	}
}

UINT16 eeprom_gen_rule(rc_rule_t* rc_rule) {
	UINT16 eeprom_address = rules_first + (rules_used++ * sizeof(rule_t));
	rule_t* rule = (rule_t*) malloc_zero(sizeof(rule_t));
	int rule_conditions_count = g_list_length(rc_rule->conditions);
	int rule_statements_count = g_list_length(rc_rule->statements);

	if(debug)
		fprintf(stderr, "Rule \"%s\" at 0x%x with %d condition(s) and %d statements(s):\n", rc_rule->identifier, eeprom_address, rule_conditions_count, rule_statements_count);

	if(rule_conditions_count) {
		rule->conditions_first = conditions_first + (conditions_used * sizeof(condition_t));
		rule->conditions_last = rule->conditions_first + ((rule_conditions_count - 1) * sizeof(condition_t));
	} else {
		rule->conditions_first = RLS_FIELD_UNUSED;
		rule->conditions_last = RLS_FIELD_UNUSED;
	}
	if(optimize_bytecode && rule_conditions_count) {
		// Optimize bytecode by reusing a previously written
		// list of items that matches the list created by the
		// current item as one of its subitems.
		accumulator_t* accumulator = malloc_zero(sizeof(accumulator_t));
		accumulator->buffer = malloc_zero(rule_conditions_count * sizeof(condition_t));
		accumulator->counter = 0;
		g_list_foreach(rc_rule->conditions, eeprom_foreach_gen_condition, accumulator);
		void* addr = memfind(accumulator->buffer, eeprom_image + conditions_first, eeprom_image + conditions_first + (conditions_used * sizeof(condition_t)), sizeof(condition_t), rule_conditions_count);
		if(addr) {
			rule->conditions_first = (UINT16) (addr - ((void*) eeprom_image));
			rule->conditions_last = rule->conditions_first + ((rule_conditions_count - 1) * sizeof(condition_t));
		} else {
			memcpy(eeprom_image + conditions_first + (conditions_used * sizeof(condition_t)), accumulator->buffer, rule_conditions_count * sizeof(condition_t));
			conditions_used += rule_conditions_count;
		}
		free(accumulator->buffer);
		free(accumulator);
	} else {
		g_list_foreach(rc_rule->conditions, eeprom_foreach_gen_condition, NULL);
	}

	if(rule_statements_count) {
		rule->statements_first = statements_first + (statements_used * sizeof(statement_t));
		rule->statements_last = rule->statements_first + ((rule_statements_count - 1) * sizeof(statement_t));
	} else {
		rule->statements_first = RLS_FIELD_UNUSED;
		rule->statements_last = RLS_FIELD_UNUSED;
	}
	if(optimize_bytecode && rule_statements_count) {
		// Optimize bytecode by reusing a previously written
		// list of items that matches the list created by the
		// current item as one of its subitems.
		accumulator_t* accumulator = malloc_zero(sizeof(accumulator_t));
		accumulator->buffer = malloc_zero(rule_statements_count * sizeof(statement_t));
		accumulator->counter = 0;
		g_list_foreach(rc_rule->statements, eeprom_foreach_gen_statement, accumulator);
		void* addr = memfind(accumulator->buffer, eeprom_image + statements_first, eeprom_image + statements_first + (statements_used * sizeof(statement_t)), sizeof(statement_t), rule_statements_count);
		if(addr) {
			rule->statements_first = (UINT16) (addr - ((void*) eeprom_image));
			rule->statements_last = rule->statements_first + ((rule_statements_count - 1) * sizeof(statement_t));
		} else {
			memcpy(eeprom_image + statements_first + (statements_used * sizeof(statement_t)), accumulator->buffer, rule_statements_count * sizeof(statement_t));
			statements_used += rule_statements_count;
		}
		free(accumulator->buffer);
		free(accumulator);
	} else {
		g_list_foreach(rc_rule->statements, eeprom_foreach_gen_statement, NULL);
	}

	memcpy(eeprom_image + eeprom_address, rule, sizeof(rule_t));
	free(rule);
	return eeprom_address;
}

void eeprom_foreach_gen_rule(gpointer data, gpointer user_data) {
	eeprom_gen_rule((rc_rule_t*) data);
}

UINT16 eeprom_gen_property(rc_property_t* rc_property) {
	UINT16 eeprom_address = properties_first + (properties_used++ * sizeof(property_t));
	property_t* property = (property_t*) malloc_zero(sizeof(property_t));

	property->key = eeprom_gen_identifier(rc_property->key);
	eeprom_gen_variable_at(rc_property->variable, &property->variable);

	memcpy(eeprom_image + eeprom_address, property, sizeof(property_t));
	free(property);
	return eeprom_address;
}

void eeprom_foreach_gen_property(gpointer data, gpointer user_data) {
	eeprom_gen_property((rc_property_t*) data);
}

UINT16 eeprom_gen_fact(rc_fact_t* rc_fact) {
	UINT16 eeprom_address = facts_first + (facts_used++ * sizeof(fact_t));
	fact_t* fact = (fact_t*) malloc_zero(sizeof(fact_t));
	int fact_properties_count = g_list_length(rc_fact->properties);

	MEM_TO_UINT16(eeprom_image, SCATTERWEB_CONFIGURATION_ID, &fact->mac_id);
	fact->name = eeprom_gen_name(rc_fact->name);

	if(fact_properties_count) {
		fact->properties_first = properties_first + (properties_used * sizeof(property_t));
		fact->properties_last = fact->properties_first + ((fact_properties_count - 1) * sizeof(property_t));
	} else {
		fact->properties_first = RLS_FIELD_UNUSED;
		fact->properties_last = RLS_FIELD_UNUSED;
	}
	g_list_foreach(rc_fact->properties, eeprom_foreach_gen_property, NULL);

	fact->timeStamp.secs = 0;
	fact->timeStamp.millis = 0;
	fact->state = RLS_FACT_MODIFIED;

	memcpy(eeprom_image + eeprom_address, fact, sizeof(fact_t));
	free(fact);
	return eeprom_address;
}

void eeprom_foreach_gen_fact(gpointer data, gpointer user_data) {
	eeprom_gen_fact((rc_fact_t*) data);
}

void eeprom_foreach_gen_depends(gpointer data, gpointer user_data) {
	fprintf(stderr, "WARNING: Ruleset interdependencies (\"depends %s\") are not supported when generating bytecode.\n", (char*) data);
}

gint by_priority(gconstpointer a, gconstpointer b) {
	rc_rule_t* rc_rule_a = (rc_rule_t*) a;
	rc_rule_t* rc_rule_b = (rc_rule_t*) b;
	if(rc_rule_a->priority > rc_rule_b->priority)
		return -1;
	if(rc_rule_a->priority < rc_rule_b->priority)
		return 1;
	return 0;
}

// Initialize global variables that define the layout of the bytecode. This is
// a separate function because the knowledge about the parameters may change as
// the bytecode is optimized.
void eeprom_init_bytecode_layout(int rules_count, int conditions_count, int statements_count, int slots_count, int expressions_count, int initializers_count, int const_variables_count, int facts_count, int properties_count) {

	// Calculate how much static space the entities of the ruleset will
	// need. Also initialize entity counters.

	rules_first = RLS_HEADER_ADDR + sizeof(rls_header_t);
	rules_last = rules_first + (rules_count - 1) * sizeof(rule_t);
	rules_used = 0;
	if(debug)
		fprintf(stderr, "rules_first:\t\t0x%x,\trules_last:\t\t0x%x,\trules_count:\t\t%2d (%3d B)\n", rules_first, rules_last, rules_count, rules_count * sizeof(rule_t));


	conditions_first = rules_last + sizeof(rule_t);
	conditions_last = conditions_first + (conditions_count - 1) * sizeof(condition_t);
	conditions_used = 0;
	if(debug)
		fprintf(stderr, "conditions_first:\t0x%x,\tconditions_last:\t0x%x,\tconditions_count:\t%2d (%3d B)\n", conditions_first, conditions_last, conditions_count, conditions_count * sizeof(condition_t));

	statements_first = conditions_last + sizeof(condition_t);
	statements_last = statements_first + (statements_count - 1) * sizeof(statement_t);
	statements_used = 0;
	if(debug)
		fprintf(stderr, "statements_first:\t0x%x,\tstatements_last:\t0x%x,\tstatements_count:\t%2d (%3d B)\n", statements_first, statements_last, statements_count, statements_count * sizeof(statement_t));

	slots_first = statements_last + sizeof(statement_t);
	slots_last = slots_first + (slots_count - 1) * sizeof(slot_t);
	slots_used = 0;
	if(debug)
		fprintf(stderr, "slots_first:\t\t0x%x,\tslots_last:\t\t0x%x,\tslots_count:\t\t%2d (%3d B)\n", slots_first, slots_last, slots_count, slots_count * sizeof(slot_t));

	expressions_first = slots_last + sizeof(slot_t);
	expressions_last = expressions_first + (expressions_count - 1) * sizeof(expression_t);
	expressions_used = 0;
	if(debug)
		fprintf(stderr, "expressions_first:\t0x%x,\texpressions_last:\t0x%x,\texpressions_count:\t%2d (%3d B)\n", expressions_first, expressions_last, expressions_count, expressions_count * sizeof(expression_t));

	initializers_first = expressions_last + sizeof(expression_t);
	initializers_last = initializers_first + (initializers_count - 1) * sizeof(initializer_t);
	initializers_used = 0;
	if(debug)
		fprintf(stderr, "initializers_first:\t0x%x,\tinitializers_last:\t0x%x,\tinitializers_count:\t%2d (%3d B)\n", initializers_first, initializers_last, initializers_count, initializers_count * sizeof(initializer_t));

	const_variables_first = initializers_last + sizeof(initializer_t);;
	const_variables_last = const_variables_first + (const_variables_count - 1) * sizeof(variable_t);
	const_variables_used = 0;
	if(debug)
		fprintf(stderr, "const_variables_first:\t0x%x,\tconst_variables_last:\t0x%x,\tconst_variables_count:\t%2d (%3d B)\n", const_variables_first, const_variables_last, const_variables_count, const_variables_count * sizeof(variable_t));

	// Facts are dynamic at runtime and hence are the last item to leave
	// room to grow upwards.
	facts_first = const_variables_last + sizeof(variable_t);
	if(facts_count) {
		facts_last = facts_first + (facts_count - 1) * sizeof(fact_t);
	} else {
		facts_last = RLS_FIELD_UNUSED;
	}
	facts_used = 0;
	if(debug)
		fprintf(stderr, "facts_first:\t\t0x%x,\tfacts_last:\t\t0x%x,\tfacts_count:\t\t%2d (%3d B)\n", facts_first, facts_last, facts_count, facts_count * sizeof(fact_t));

	// Just as facts, properties are dynamic at runtime. They grow
	// downwards in the memory space.
	if(properties_count) {
		properties_first = RLS_MAX_ADDR - (properties_count * sizeof(property_t)) + 1;
	} else {
		properties_first = RLS_FIELD_UNUSED;
	}
	properties_last = RLS_MAX_ADDR - sizeof(property_t);
	properties_used = 0;
	if(debug)
		fprintf(stderr, "properties_first:\t0x%x,\tproperties_last:\t0x%x,\tproperties_count:\t%2d (%3d B)\n", properties_first, properties_last, properties_count, properties_count * sizeof(property_t));

	if(debug)
		fprintf(stderr, "Total bytecode size: %d B\n",
			sizeof(UINT16) +
			sizeof(rls_header_t) +
			(rules_count * sizeof(rule_t)) +
			(conditions_count * sizeof(condition_t)) +
			(statements_count * sizeof(statement_t)) +
			(slots_count * sizeof(slot_t)) +
			(expressions_count * sizeof(expression_t)) +
			(initializers_count * sizeof(initializer_t)) +
			(const_variables_count * sizeof(variable_t)) +
			(facts_count * sizeof(fact_t)) +
			(properties_count * sizeof(property_t))
		);
}

void eeprom_gen_ruleset(rc_ruleset_t* rc_ruleset) {
	cur_ruleset = rc_ruleset;

	// Sort rules by priority to make processing easier for the rule engine.
	rc_ruleset->rules = g_list_sort(rc_ruleset->rules, by_priority);

	// Get a reference to the location of the FACTS bytecode header.
	rls_header_t* rls_header = (rls_header_t*) (eeprom_image + RLS_HEADER_ADDR);

	if(debug) {
		fprintf(stderr, "Generating bytcode for ruleset with %d rules, %d conditions, %d statements, %d slots, %d expressions, %d initializers, %d variables, %d facts, and %d properties.\n\n", rc_ruleset->rules_count, rc_ruleset->conditions_count, rc_ruleset->statements_count, rc_ruleset->slots_count, rc_ruleset->expressions_count, rc_ruleset->initializers_count, rc_ruleset->const_variables_count, rc_ruleset->facts_count, rc_ruleset->properties_count);
		fprintf(stderr, "EEPROM bytcode generator run #1...\n");
	}

	// Initialize the layout of the bytecode.
	eeprom_init_bytecode_layout(rc_ruleset->rules_count, rc_ruleset->conditions_count, rc_ruleset->statements_count, rc_ruleset->slots_count, rc_ruleset->expressions_count, rc_ruleset->initializers_count, rc_ruleset->const_variables_count, rc_ruleset->facts_count, rc_ruleset->properties_count);

	g_list_foreach(rc_ruleset->depends, eeprom_foreach_gen_depends, NULL);

	// Write the ruleset to EEPROM image.
	g_list_foreach(rc_ruleset->rules, eeprom_foreach_gen_rule, NULL);
	g_list_foreach(rc_ruleset->facts, eeprom_foreach_gen_fact, NULL);
	if(debug)
		fprintf(stderr, "...DONE.\n\n");

	if(optimize_bytecode) {

		// In order to optimize the bytecode, we used to data gathered
		// during the previous run to recalculate the layout of the
		// bytecode image and then simply write it again.

		if(debug)
			fprintf(stderr, "EEPROM bytcode generator run #2...\n");

		// First clear the data of the previous run from the EEPROM image.
		memset(eeprom_image + rules_first, 0, (const_variables_last + sizeof(variable_t)) - rules_first);
		if(facts_used)
			memset(eeprom_image + facts_first, 0, (facts_last + sizeof(fact_t)) - facts_first);
		if(properties_used)
			memset(eeprom_image + properties_first, 0, (properties_last + sizeof(property_t)) - properties_first);

		// Re-initialize the layout of the bytecode using the new data.
		eeprom_init_bytecode_layout(rules_used, conditions_used, statements_used, slots_used, expressions_used, initializers_used, const_variables_used, facts_used, properties_used);

		// Re-write the optimized ruleset to EEPROM image.
		g_list_foreach(rc_ruleset->rules, eeprom_foreach_gen_rule, NULL);
		g_list_foreach(rc_ruleset->facts, eeprom_foreach_gen_fact, NULL);
		if(debug)
			fprintf(stderr, "...DONE.\n\n");
	}

	// Write the header.
	rls_header->rules_first = rules_first;
	rls_header->rules_last = rules_last;
	rls_header->facts_first = facts_first;
	rls_header->facts_last = facts_last;
	rls_header->properties_first = properties_first;
	rls_header->properties_last = properties_last;

	// Sanity check: Facts and their properties must not overlap.
	if(
		(
			(rls_header->facts_first != RLS_FIELD_UNUSED)
			|| (rls_header->properties_first != RLS_FIELD_UNUSED)
		)
		&& (rls_header->facts_last > rls_header->properties_first)
	) {
		fprintf(stderr, "EEPROM too small for ruleset.\n");
		exit(RC_EEEPROM);
	}
}

void eeprom_foreach_gen_ruleset(gpointer data, gpointer user_data) {
	eeprom_gen_ruleset((rc_ruleset_t*) data);
}

void eeprom_generate(GList* rulesets, FILE* eeprom_file) {

	// Allocate space for the EEPROM image and zero it.
	eeprom_image = malloc_zero(RLS_MAX_ADDR + 1);

	// Initialize first part of the EEPROM image with the vanilla EEPROM
	// file.
	if(fread(eeprom_image, 1, RLS_MAGIC_ADDR, eeprom_file) != RLS_MAGIC_ADDR) {
		fprintf(stderr, "Input EEPROM file too short.\n");
		exit(RC_EEEPROM);
	}

	// Write FACTS magic number to EEPROM image.
	UINT16_TO_MEM(RLS_MAGIC_NUMBER, eeprom_image, RLS_MAGIC_ADDR);

	// Initialize identifiers hash table.
	identifiers = g_hash_table_new(g_str_hash, g_str_equal);
	identifier_count = RLS_SLOT_NAME_THIS;	// identifier_count is incremented before each use, so RLS_SLOT_NAME_THIS is never used.

	// Write ruleset to EEPROM image.
	g_list_foreach(rulesets, eeprom_foreach_gen_ruleset, NULL);

	// Free identifiers hash table.
	g_hash_table_destroy(identifiers);

	// Write EEPROM image to file.
	if(fwrite(eeprom_image, 1, (RLS_MAX_ADDR + 1), yyout) != (RLS_MAX_ADDR + 1)) {
		fprintf(stderr, "Error while writing to EEPROM file.\n");
		exit(RC_EIO);
	}

	// Free EEPROM image.
	free(eeprom_image);
}
