%{

/*
 * Copyright 2005, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * Contributors: Georg Wittenburg
 */

#include "main.h"

/**
 * This file is used by yacc to generate the parser.
 **/

// These are temporary variables used when a non-terminal symbol is parsed into
// a list structure, i.e. contains multiple items.
GList* cur_depends;
GList* cur_facts;
GList* cur_rules;
GList* cur_functions;
GList* cur_slots;
GList* cur_names;

/*
 * cur_conditions is special as it may be used in recursive calls: Conditions
 * may contain slots which again may contain conditions, etc. Hence one needs
 * to take care that save the previous result somewhere else when parsing.
 * If cur_conditions is NULL, then nothing needs to be saved. Otherwise the
 * current value is pushed into cur_conditions_stack and restored later on.
 */
GList* cur_conditions;
GList* cur_conditions_stack;

GList* cur_statements;
GList* cur_initializers;
GList* cur_expressions;
GList* cur_properties;

// These are counters on how many items of a particular type have been parsed.
int cur_facts_count;
int cur_rules_count;
int cur_functions_count;
int cur_slots_count;
int cur_names_count;
int cur_expressions_count;
int cur_conditions_count;
int cur_statements_count;
int cur_initializers_count;
int cur_const_variables_count;
int cur_properties_count;

/*
 * As mentioned above, conditions may be parsed recursively and thus need
 * special attention.
 */
void save_cur_conditions() {
	if(debug)
		fprintf(stderr, "    -> Pushing %d condition(s). %d conditions saved on stack.\n", g_list_length(cur_conditions), g_list_length(cur_conditions_stack));
	cur_conditions_stack = g_list_prepend(cur_conditions_stack, cur_conditions);
	cur_conditions = NULL;
}

void restore_cur_conditions() {
	if(g_list_length(cur_conditions_stack) == 0) {
		yyerror("INTERNAL ERROR: Trying to pop empty stack.");
		exit(RC_EINTERNAL);
	}
	GList* elem = g_list_first(cur_conditions_stack);
	cur_conditions = elem->data;
	cur_conditions_stack = g_list_remove(cur_conditions_stack, cur_conditions);
	if(debug)
		fprintf(stderr, "    -> Popped %d condition(s). %d conditions saved on stack.\n", g_list_length(cur_conditions), g_list_length(cur_conditions_stack));
}

/*
 * Slots and names may be defined explicitly with an identifier. These two
 * functions retrieve the actual struct based on the identifier.
 */
rc_slot_t* retrieve_slot(char* identifier) {
	rc_slot_t* slot = NULL;
	GList* elem = g_list_first(cur_slots);
	while(elem != NULL) {
		slot = (rc_slot_t*) elem->data;
		if(strcmp(slot->identifier, identifier) == 0)
			break;
		else
			slot = NULL;
		elem = g_list_next(elem);
	}
	return slot;
}

rc_name_t* retrieve_name(char* identifier) {
	if(strcmp(identifier, rc_name_this.identifier) == 0)
		return &rc_name_this;
	rc_name_t* name = NULL;
	GList* elem = g_list_first(cur_names);
	while(elem != NULL) {
		name = (rc_name_t*) elem->data;
		if(strcmp(name->identifier, identifier) == 0)
			break;
		else
			name = NULL;
		elem = g_list_next(elem);
	}
	return name;
}

%}

// Return types of productions.
%union
{
	int			int_t;
	char*			str_t;
	rc_fact_t*		fact_t;
	rc_rule_t*		rule_t;
	rc_slot_t*		slot_t;
	rc_name_t*		name_t;
	rc_variable_t*		variable_t;
	rc_condition_t*		condition_t;
	rc_statement_t*		statement_t;
	rc_expression_t*	expression_t;
}

// Start symbol.
%start ruleset

// Terminal symbols. See rc.lex for declarations.
%token RC_ASSIGN
%token RC_OPEN_ROUND_BRACKET RC_CLOSE_ROUND_BRACKET
%token RC_OPEN_SQUARE_BRACKET RC_CLOSE_SQUARE_BRACKET
%token RC_OPEN_CURLY_BRACKET RC_CLOSE_CURLY_BRACKET
%token RC_RIGHT_ARROW RC_LEFT_ARROW RC_COMMA RC_QUOTES
%token RC_RULESET RC_DEPENDS  RC_NAME  RC_SLOT RC_FACT RC_RULE
%token RC_EXISTS RC_EVAL
%token RC_DEFINE RC_RETRACT RC_COPY RC_SEND RC_SET RC_FLUSH RC_CALL
%token RC_EQUALS RC_NOT_EQUALS RC_LESS_THAN RC_GREATER_THAN RC_LESS_OR_EQUAL RC_GREATER_OR_EQUAL
%token RC_COUNT RC_SUM RC_PRODUCT RC_MINIMUM RC_MAXIMUM RC_SQUARE_ROOT
%token RC_ADD RC_SUBTRACT RC_MULTIPLY RC_DIVIDE RC_TO_THE_POWER_OF
%token <int_t> RC_BOOL
%token <int_t> RC_INTEGER
%token <str_t> RC_STRING

%%

ruleset:
	RC_RULESET identifier {
		// Initialize all global temporary variables. They are updated
		// while parsing.
		cur_depends = NULL;
		cur_facts = NULL;
		cur_facts_count = 0;
		cur_rules = NULL;
		cur_rules_count = 0;
		cur_functions = NULL;
		cur_functions_count = 0;
		cur_slots = NULL;
		cur_slots_count = 0;
		cur_names = NULL;
		cur_names_count = 0;
		cur_expressions_count = 0;
		cur_conditions_count = 0;
		cur_statements_count = 0;
		cur_initializers_count = 0;
		cur_const_variables_count = 0;
		cur_properties_count = 0;
		cur_conditions_stack = NULL;
	} block_list {
		rc_ruleset_t* ruleset = (rc_ruleset_t*) malloc_zero(sizeof(rc_ruleset_t));
		ruleset->identifier = $<str_t>2;
		ruleset->depends = cur_depends;
		ruleset->facts = cur_facts;
		ruleset->facts_count = cur_facts_count;
		ruleset->rules = cur_rules;
		ruleset->rules_count = cur_rules_count;
		ruleset->functions = cur_functions;
		ruleset->functions_count = cur_functions_count;
		ruleset->slots = cur_slots;
		ruleset->slots_count = cur_slots_count;
		ruleset->names = cur_names;
		ruleset->names_count = cur_names_count;
		ruleset->expressions_count = cur_expressions_count;
		ruleset->conditions_count = cur_conditions_count;
		ruleset->statements_count = cur_statements_count;
		ruleset->initializers_count = cur_initializers_count;
		ruleset->const_variables_count = cur_const_variables_count;
		ruleset->properties_count = cur_properties_count;
		rulesets = g_list_append(rulesets, ruleset);
	}
;

identifier:
	RC_STRING
;

block_list:
	block
	| block block_list
;

block:
	depends {
		cur_depends = g_list_append(cur_depends, $<str_t>1);
	}
	| named_name {
		cur_names = g_list_append(cur_names, $<name_t>1);
	}
	| named_slot {
		cur_slots = g_list_append(cur_slots, $<slot_t>1);
	}
	| fact {
		cur_facts = g_list_append(cur_facts, $<fact_t>1);
	}
	| rule {
		cur_rules = g_list_append(cur_rules, $<rule_t>1);
	}
;

depends:
	RC_DEPENDS identifier {
		$<str_t>$ = $<str_t>2;
	}
;

named_name:
	RC_NAME identifier RC_ASSIGN name {
		rc_name_t* name = $<name_t>4;
		name->identifier = $<str_t>2;
		$<name_t>$ = name;
	}
;

named_slot:
	RC_SLOT identifier RC_ASSIGN slot {
		rc_slot_t* slot = $<slot_t>4;
		slot->identifier = $<str_t>2;
		$<slot_t>$ = slot;
	}
;

fact:
	RC_FACT name {
		cur_properties = NULL;
	} property_list_opt {
		rc_fact_t* fact = (rc_fact_t*) malloc_zero(sizeof(rc_fact_t));
		cur_facts_count++;
		fact->name = $<name_t>2;
		fact->properties = cur_properties;
		fact->unique_id = cur_facts_count;
		$<fact_t>$ = fact;
	}
;

rule:
	RC_RULE identifier priority {
		if(cur_conditions != NULL) {
			yyerror("INTERNAL ERROR: Some conditions are left unprocessed from previous definitions.");
			exit(RC_EINTERNAL);
		}
		if(debug)
			fprintf(stderr, "Parsing rule conditions...\n");
	} condition_list {
		cur_statements = NULL;
		if(debug)
			fprintf(stderr, "Parsing rule statements...\n");
	} statement_list {
		rc_rule_t* rule = (rc_rule_t*) malloc_zero(sizeof(rc_rule_t));
		cur_rules_count++;
		rule->identifier = $<str_t>2;
		rule->priority = $<int_t>3;
		rule->conditions = cur_conditions;
		rule->statements = cur_statements;
		cur_conditions = NULL;
		if(debug)
			fprintf(stderr, "...parsed rule \"%s\" with %d condition(s) and %d statement(s).\n", rule->identifier, g_list_length(rule->conditions), g_list_length(rule->statements));
		$<rule_t>$ = rule;
	}
;

priority:
	RC_INTEGER
;

slot:
	identifier {
		rc_slot_t* slot = retrieve_slot($<str_t>1);
		if(slot == NULL)
			yyerror("Slot not defined.");
		if(debug)
			fprintf(stderr, "    Used slot \"%s\", \"%s\" with %d condition(s).\n", slot->name->string, slot->key, g_list_length(slot->conditions));
		$<slot_t>$ = slot;
	}
	| RC_OPEN_CURLY_BRACKET name {
		if(debug)
			fprintf(stderr, "    Parsing slot...\n");
		save_cur_conditions();
	} condition_list_opt RC_CLOSE_CURLY_BRACKET {
		rc_slot_t* slot = malloc_zero(sizeof(rc_slot_t));
		cur_slots_count++;
		slot->identifier = NULL;
		slot->name = $<name_t>2;
		slot->key = NULL;
		slot->conditions = cur_conditions;
		restore_cur_conditions();
		if(debug)
			fprintf(stderr, "    Parsed slot \"%s\", \"%s\" with %d condition(s).\n", slot->name->string, slot->key, g_list_length(slot->conditions));
		$<slot_t>$ = slot;
	}
	| RC_OPEN_CURLY_BRACKET name key {
		if(debug)
			fprintf(stderr, "    Parsing slot...\n");
		save_cur_conditions();
	} condition_list_opt RC_CLOSE_CURLY_BRACKET {
		rc_slot_t* slot = malloc_zero(sizeof(rc_slot_t));
		cur_slots_count++;
		slot->identifier = NULL;
		slot->name = $<name_t>2;
		slot->key = $<str_t>3;
		slot->conditions = cur_conditions;
		restore_cur_conditions();
		if(debug)
			fprintf(stderr, "    Parsed slot \"%s\", \"%s\" with %d condition(s).\n", slot->name->string, slot->key, g_list_length(slot->conditions));
		$<slot_t>$ = slot;
	}
;

condition_list_opt:
	/* empty */
	| condition_list

condition_list:
	RC_LEFT_ARROW condition {
		cur_conditions = g_list_append(cur_conditions, $<condition_t>2);
	}
	| RC_LEFT_ARROW condition {
		cur_conditions = g_list_append(cur_conditions, $<condition_t>2);
	} condition_list
;

condition:
	RC_EXISTS slot {
		rc_condition_t* condition = (rc_condition_t*) malloc_zero(sizeof(rc_condition_t));
		cur_conditions_count++;
		condition->type = RC_EXISTS;
		condition->exists_slot = $<slot_t>2;
		if(debug)
			fprintf(stderr, "  Parsed EXISTS condition.\n");
		$<condition_t>$ = condition;
	}
	| RC_EVAL RC_OPEN_ROUND_BRACKET expression comparison_operation expression RC_CLOSE_ROUND_BRACKET {
		rc_condition_t* condition = (rc_condition_t*) malloc_zero(sizeof(rc_condition_t));
		cur_conditions_count++;
		condition->type = RC_EVAL;
		condition->eval_left_expression = $<expression_t>3;
		condition->eval_comparison_operation = $<int_t>4;
		condition->eval_right_expression = $<expression_t>5;
		if(debug)
			fprintf(stderr, "  Parsed EVAL condition.\n");
		$<condition_t>$ = condition;
	}
;

comparison_operation:
	RC_EQUALS {$<int_t>$ = RC_EQUALS;}
	| RC_NOT_EQUALS {$<int_t>$ = RC_NOT_EQUALS;}
	| RC_LESS_THAN {$<int_t>$ = RC_LESS_THAN;}
	| RC_GREATER_THAN {$<int_t>$ = RC_GREATER_THAN;}
	| RC_LESS_OR_EQUAL {$<int_t>$ = RC_LESS_OR_EQUAL;}
	| RC_GREATER_OR_EQUAL {$<int_t>$ = RC_GREATER_OR_EQUAL;}
;

statement_list:
	RC_RIGHT_ARROW statement {
		cur_statements = g_list_append(cur_statements, $<statement_t>2);
	}
	| RC_RIGHT_ARROW statement {
		cur_statements = g_list_append(cur_statements, $<statement_t>2);
	} statement_list
;

statement:
	RC_DEFINE name {
		cur_initializers = NULL;
	} initializer_list_opt {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_DEFINE;
		statement->define_name = $<name_t>2;
		statement->define_initializers = cur_initializers;
		if(debug)
			fprintf(stderr, "  Parsed DEFINE statement.\n");
		$<statement_t>$ = statement;
	}
	| RC_RETRACT slot {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_RETRACT;
		statement->retract_slot = $<slot_t>2;
		if(debug)
			fprintf(stderr, "  Parsed RETRACT statement.\n");
		$<statement_t>$ = statement;
	}
	| RC_COPY slot {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_COPY;
		statement->copy_slot = $<slot_t>2;
		if(debug)
			fprintf(stderr, "  Parsed COPY statement.\n");
		$<statement_t>$ = statement;
	}
	| RC_SEND expression expression slot {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_SEND;
		statement->send_address_expression = $<expression_t>2;
		statement->send_power_expression = $<expression_t>3;
		statement->send_slot = $<slot_t>4;
		if(debug)
			fprintf(stderr, "  Parsed SEND statement.\n");
		$<statement_t>$ = statement;
	}
	| RC_SET slot RC_ASSIGN expression {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_SET;
		statement->set_slot = $<slot_t>2;
		statement->set_expression = $<expression_t>4;
		if(debug)
			fprintf(stderr, "  Parsed SET statement.\n");
		$<statement_t>$ = statement;
	}
	| RC_FLUSH slot {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_FLUSH;
		statement->flush_slot = $<slot_t>2;
		if(debug)
			fprintf(stderr, "  Parsed FLUSH statement.\n");
		$<statement_t>$ = statement;
	}
	| RC_CALL identifier {
		cur_expressions = NULL;
	} expression_list_opt {
		rc_statement_t* statement = (rc_statement_t*) malloc_zero(sizeof(rc_statement_t));
		cur_statements_count++;
		statement->type = RC_CALL;
		statement->call_identifier = $<str_t>2;
		statement->call_expressions = cur_expressions;
		if(debug)
			fprintf(stderr, "  Parsed CALL statement.\n");
		$<statement_t>$ = statement;
	}
;

expression_list_opt:
	/* empty */
	|  RC_OPEN_ROUND_BRACKET expression_list RC_CLOSE_ROUND_BRACKET
;

expression_list:
	expression {
		cur_expressions = g_list_append(cur_expressions, $<expression_t>1);
	}
	| expression {
		cur_expressions = g_list_append(cur_expressions, $<expression_t>1);
	} RC_COMMA expression_list
;

expression:
	variable {
		rc_expression_t* expression = malloc_zero(sizeof(rc_expression_t));
		cur_expressions_count++;
		cur_const_variables_count++;
		expression->type = RC_CONSTANT;
		expression->constant_variable = $<variable_t>1;
		$<expression_t>$ = expression;
	}
	| RC_OPEN_ROUND_BRACKET unary_operation expression RC_CLOSE_ROUND_BRACKET {
		rc_expression_t* expression = malloc_zero(sizeof(rc_expression_t));
		cur_expressions_count++;
		expression->type = RC_UNARY_OPERATION;
		expression->unary_operation = $<int_t>2;
		expression->unary_expression = $<expression_t>3;
		$<expression_t>$ = expression;
	}
	| RC_OPEN_ROUND_BRACKET expression binary_operation expression RC_CLOSE_ROUND_BRACKET {
		rc_expression_t* expression = malloc_zero(sizeof(rc_expression_t));
		cur_expressions_count++;
		expression->type = RC_BINARY_OPERATION;
		expression->left_binary_expression = $<expression_t>2;
		expression->binary_operation = $<int_t>3;
		expression->right_binary_expression = $<expression_t>4;
		$<expression_t>$ = expression;
	}
;

unary_operation:
	RC_COUNT {$<int_t>$ = RC_COUNT;}
	| RC_SUM {$<int_t>$ = RC_SUM;}
	| RC_PRODUCT {$<int_t>$ = RC_PRODUCT;}
	| RC_MINIMUM {$<int_t>$ = RC_MINIMUM;}
	| RC_MAXIMUM {$<int_t>$ = RC_MAXIMUM;}
	| RC_SQUARE_ROOT {$<int_t>$ = RC_SQUARE_ROOT;}
;

binary_operation:
	RC_ADD {$<int_t>$ = RC_ADD;}
	| RC_SUBTRACT {$<int_t>$ = RC_SUBTRACT;}
	| RC_MULTIPLY {$<int_t>$ = RC_MULTIPLY;}
	| RC_DIVIDE {$<int_t>$ = RC_DIVIDE;}
	| RC_TO_THE_POWER_OF {$<int_t>$ = RC_TO_THE_POWER_OF;}
;

property_list_opt:
	/* empty */
	|  RC_OPEN_SQUARE_BRACKET property_list RC_CLOSE_SQUARE_BRACKET
;

property_list:
	property
	| property RC_COMMA property_list
;

property:
	key RC_ASSIGN variable {
		rc_property_t* property = (rc_property_t*) malloc_zero(sizeof(rc_property_t));
		cur_properties_count++;
		property->key = $<str_t>1;
		property->variable = $<variable_t>3;
		cur_properties = g_list_append(cur_properties, property);
	}
;

key:
	RC_STRING
;

variable:
	RC_BOOL {
		rc_variable_t* variable = (rc_variable_t*) malloc_zero(sizeof(rc_variable_t));
		variable->type = RC_BOOL;
		variable->value = (void*) $1;
		$<variable_t>$ = variable;
	}
	| RC_INTEGER {
		rc_variable_t* variable = (rc_variable_t*) malloc_zero(sizeof(rc_variable_t));
		variable->type = RC_INTEGER;
		variable->value = (void*) $1;
		$<variable_t>$ = variable;
	}
	| quoted_string {
		rc_variable_t* variable = (rc_variable_t*) malloc_zero(sizeof(rc_variable_t));
		variable->type = RC_STRING;
		variable->value = $<str_t>1;
		$<variable_t>$ = variable;
	}
	| slot {
		rc_variable_t* variable = (rc_variable_t*) malloc_zero(sizeof(rc_variable_t));
		variable->type = RC_SLOT;
		variable->value = $<slot_t>1;
		$<variable_t>$ = variable;
	}
;

initializer_list_opt:
	/* empty */
	|  RC_OPEN_SQUARE_BRACKET initializer_list RC_CLOSE_SQUARE_BRACKET
;

initializer_list:
	initializer
	| initializer RC_COMMA initializer_list
;

initializer:
	key RC_ASSIGN expression {
		rc_initializer_t* initializer = (rc_initializer_t*) malloc_zero(sizeof(rc_initializer_t));
		cur_initializers_count++;
		initializer->key = $<str_t>1;
		initializer->expression = $<expression_t>3;
		cur_initializers = g_list_append(cur_initializers, initializer);
	}
;

name:
	identifier {
		rc_name_t* name = retrieve_name($<str_t>1);
		if(name == NULL)
			yyerror("Name not defined.");
		$<name_t>$ = name;
	}
	| quoted_string {
		if(strcmp($<str_t>1, rc_name_this.string) == 0) {
			yyerror("ERROR: \"this\" is a reserved name.");
			exit(RC_ECOMPILE);
		}
		rc_name_t* name = malloc_zero(sizeof(rc_name_t));
		cur_names_count++;
		name->identifier = NULL;
		name->string = $<str_t>1;
		$<name_t>$ = name;
	}
;

quoted_string:
	RC_QUOTES RC_QUOTES {
		$<str_t>$ = strdup("");
	}
	| RC_QUOTES RC_STRING RC_QUOTES {
		$<str_t>$ = $2;
	}
;

%%

void yyerror(char* msg) {
	errors++;
	fprintf(stderr, "%s:%d: %s Current token is \"%s\"\n", in_filename, line, msg, yytext);
}
