/*
 * Copyright 2005, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * Contributors: Georg Wittenburg
 */

#include "main.h"
#include "y.tab.h"

/**
 * This file frees the memory after the work has been done.
 **/

void free_name(rc_name_t* name, int delete_named_names) {
	// Similarly to named slots, named names (which admittedly sounds
	// funny) may be used in multiple places in the code. By deleting named
	// and anonymous names seperately depending on the context, we make
	// sure that we don't free the same memory location twice.
	if(name == &rc_name_this)
		return;	// Don't free static constants.
	if(
		(delete_named_names && (name->identifier != NULL))
		|| (!delete_named_names && (name->identifier == NULL))
	) {
		if(name->identifier != NULL)
			free(name->identifier);
		free(name->string);
		free(name);
	}
}

void foreach_free_name(gpointer data, gpointer user_data) {
	free_name((rc_name_t*) data, (int) user_data);
}

void free_slot(rc_slot_t* slot, int free_named_slots);

void free_variable(rc_variable_t* variable) {
	if(variable == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL variable.\n");
		exit(RC_EINTERNAL);
	}
	switch(variable->type) {
		case RC_BOOL:
		case RC_INTEGER:
			break;
		case RC_STRING:
			free(variable->value);
			break;
		case RC_SLOT:
			free_slot((rc_slot_t*) variable->value, 0);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid variable type %d.\n", variable->type);
			exit(RC_EINTERNAL);
	}
	free(variable);
}

void foreach_free_condition(gpointer data, gpointer user_data);

void free_slot(rc_slot_t* slot, int free_named_slots) {
	// Named slots (as opposed to anonymous slots) may be used in multiple
	// places in the code. By deleting named and anonymous slots seperately
	// depending on the context, we make sure that we don't free the same
	// memory location twice.
	if(slot == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL slot.\n");
		exit(RC_EINTERNAL);
	}
	if(
		(free_named_slots && (slot->identifier != NULL))
		|| (!free_named_slots && (slot->identifier == NULL))
	) {
		if(slot->identifier != NULL) {
			free(slot->identifier);
		}
		free_name(slot->name, 0);
		if(slot->key != NULL)
			free(slot->key);
		if(slot->conditions != NULL) {
			g_list_foreach(slot->conditions, foreach_free_condition, NULL);
			g_list_free(slot->conditions);
		}
		free(slot);
	}
}

void foreach_free_slot(gpointer data, gpointer user_data) {
	free_slot((rc_slot_t*) data, (int) user_data);
}

void free_expression(rc_expression_t* expression) {
	if(expression == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL expression.\n");
		exit(RC_EINTERNAL);
	}
	switch(expression->type) {
		case RC_CONSTANT:
			free_variable(expression->constant_variable);
			break;
		case RC_UNARY_OPERATION:
			free_expression(expression->unary_expression);
			break;
		case RC_BINARY_OPERATION:
			free_expression(expression->left_binary_expression);
			free_expression(expression->right_binary_expression);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid expresssion type %d.\n", expression->type);
			exit(RC_EINTERNAL);
	}
	free(expression);
}

void foreach_free_expression(gpointer data, gpointer user_data) {
	free_expression((rc_expression_t*) data);
}

void free_condition(rc_condition_t* condition) {
	if(condition == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL condition.\n");
		exit(RC_EINTERNAL);
	}
	switch(condition->type) {
		case RC_EXISTS:
			free_slot(condition->exists_slot, 0);
			break;
		case RC_EVAL:
			free_expression(condition->eval_left_expression);
			free_expression(condition->eval_right_expression);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid condition type %d.\n", condition->type);
			exit(RC_EINTERNAL);
	}
	free(condition);
}

void foreach_free_condition(gpointer data, gpointer user_data) {
	free_condition((rc_condition_t*) data);
}

void free_initializer(rc_initializer_t* initializer) {
	if(initializer == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL initializer.\n");
		exit(RC_EINTERNAL);
	}
	free(initializer->key);
	free_expression(initializer->expression);
	free(initializer);
}

void foreach_free_initializer(gpointer data, gpointer user_data) {
	free_initializer((rc_initializer_t*) data);
}

void free_statement(rc_statement_t* statement) {
	if(statement == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL statement.\n");
		exit(RC_EINTERNAL);
	}
	switch(statement->type) {
		case RC_DEFINE:
			free_name(statement->define_name, 0);
			g_list_foreach(statement->define_initializers, foreach_free_initializer, NULL);
			break;
		case RC_RETRACT:
			free_slot(statement->retract_slot, 0);
			break;
		case RC_COPY:
			free_slot(statement->copy_slot, 0);
			break;
		case RC_SEND:
			free_expression(statement->send_address_expression);
			free_expression(statement->send_power_expression);
			free_slot(statement->send_slot, 0);
			break;
		case RC_SET:
			free_slot(statement->set_slot, 0);
			free_expression(statement->set_expression);
			break;
		case RC_FLUSH:
			free_slot(statement->flush_slot, 0);
			break;
		case RC_CALL:
			free(statement->call_identifier);
			g_list_foreach(statement->call_expressions, foreach_free_expression, NULL);
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid statement type %d.\n", statement->type);
			exit(RC_EINTERNAL);
	}
	free(statement);
}

void foreach_free_statement(gpointer data, gpointer user_data) {
	free_statement((rc_statement_t*) data);
}

void free_rule(rc_rule_t* rule) {
	if(rule == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL rule.\n");
		exit(RC_EINTERNAL);
	}
	free(rule->identifier);
	g_list_foreach(rule->conditions, foreach_free_condition, NULL);
	g_list_free(rule->conditions);
	g_list_foreach(rule->statements, foreach_free_statement, NULL);
	g_list_free(rule->statements);
	free(rule);
}

void foreach_free_rule(gpointer data, gpointer user_data) {
	free_rule((rc_rule_t*) data);
}

void free_property(rc_property_t* property) {
	if(property == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL property.\n");
		exit(RC_EINTERNAL);
	}
	free(property->key);
	free_variable(property->variable);
	free(property);
}

void foreach_free_property(gpointer data, gpointer user_data) {
	free_property((rc_property_t*) data);
}

void free_fact(rc_fact_t* fact) {
	if(fact == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL fact.\n");
		exit(RC_EINTERNAL);
	}
	free_name(fact->name, 0);
	g_list_foreach(fact->properties, foreach_free_property, NULL);
	g_list_free(fact->properties);
	free(fact);
}

void foreach_free_fact(gpointer data, gpointer user_data) {
	free_fact((rc_fact_t*) data);
}

void free_function(rc_function_t* function) {
	if(function == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL function.\n");
		exit(RC_EINTERNAL);
	}
	free(function->identifier);
	free(function);
}

void foreach_free_function(gpointer data, gpointer user_data) {
	free_function((rc_function_t*) data);
}

void foreach_free_depends(gpointer data, gpointer user_data) {
	free((char*) data);
}

void free_ruleset(rc_ruleset_t* ruleset) {
	if(ruleset == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Trying to free NULL ruleset.\n");
		exit(RC_EINTERNAL);
	}
	free(ruleset->identifier);
	g_list_foreach(ruleset->depends, foreach_free_depends, NULL);
	g_list_free(ruleset->depends);
	g_list_foreach(ruleset->facts, foreach_free_fact, NULL);
	g_list_free(ruleset->facts);
	g_list_foreach(ruleset->rules, foreach_free_rule, NULL);
	g_list_free(ruleset->rules);
	g_list_foreach(ruleset->functions, foreach_free_function, NULL);
	g_list_free(ruleset->functions);
	g_list_foreach(ruleset->slots, foreach_free_slot, (gpointer) 1);
	g_list_free(ruleset->slots);
	g_list_foreach(ruleset->names, foreach_free_name, (gpointer) 1);
	g_list_free(ruleset->names);
	free(ruleset);
}

void foreach_free_ruleset(gpointer data, gpointer user_data) {
	free_ruleset((rc_ruleset_t*) data);
}
