/*
 * Copyright 2005, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * Contributors: Georg Wittenburg
 */

#include "main.h"
#include "hs_gen.h"
#include "y.tab.h"

/**
 * This file generates Haskell code.
 **/

// Not really an iterator, but a struct that helps keeping track of where we
// are while being called from a g_list_foreach loop.
typedef struct {
	int i;
	int len;
} iterator_t;

// The rulset currently being worked on.
rc_ruleset_t* cur_ruleset;

void hs_gen_name(rc_name_t* rc_name, int use_identifier_if_possible) {
	if(use_identifier_if_possible && (rc_name->identifier != NULL))
		fprintf(yyout, "%sName", rc_name->identifier);
	else
		fprintf(yyout, "(newName namespace \"%s\")", rc_name->string);
}

void hs_gen_named_name(rc_name_t* rc_name) {
	if(rc_name->identifier == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Name of named name not initialized.\n");
		exit(RC_EINTERNAL);
	}
	fprintf(yyout, "%sName = ", rc_name->identifier);
	hs_gen_name(rc_name, 0);
	fprintf(yyout, "\n");
}

void hs_foreach_gen_named_name(gpointer data, gpointer user_data) {
	hs_gen_named_name((rc_name_t*) data);
}

void hs_gen_slot(rc_slot_t* rc_slot, int use_identifier_if_possible);

void hs_gen_variable(rc_variable_t* rc_variable) {
	switch(rc_variable->type) {
		case RC_BOOL:
			if(rc_variable->value)
				fprintf(yyout, "(varBool True)");
			else
				fprintf(yyout, "(varBool False)");
			break;
		case RC_INTEGER:
			{
				int value = (int) rc_variable->value;
				if(value >= 0)
					fprintf(yyout, "(varInt %d)", value);
				else
					fprintf(yyout, "(varInt (%d))", value);
			}
			break;
		case RC_STRING:
			fprintf(yyout, "(varString \"%s\")", (char*) rc_variable->value);
			break;
		case RC_SLOT:
			fprintf(yyout, "(varSlot ");
			hs_gen_slot((rc_slot_t*) rc_variable->value, 1);
			fprintf(yyout, ")");
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid variable type %d.\n", rc_variable->type);
			exit(RC_EINTERNAL);
	}
}

void hs_foreach_gen_condition(gpointer data, gpointer user_data);

void hs_gen_slot(rc_slot_t* rc_slot, int use_identifier_if_possible) {
	if(use_identifier_if_possible && (rc_slot->identifier != NULL)) {
		fprintf(yyout, "%sSlot", rc_slot->identifier);
	} else {
		iterator_t* iter = (iterator_t*) malloc_zero(sizeof(iterator_t));
		iter->i = 0;
		iter->len = g_list_length(rc_slot->conditions);
		fprintf(yyout, "(newSlot ");
		hs_gen_name(rc_slot->name, 1);
		if(rc_slot->key != NULL)
			fprintf(yyout, " \"%s\" [", rc_slot->key);
		else
			fprintf(yyout, " \"\" [");
		g_list_foreach(rc_slot->conditions, hs_foreach_gen_condition, iter);
		fprintf(yyout, "])");
		free(iter);
	}
}

void hs_gen_named_slot(rc_slot_t* rc_slot) {
	if(rc_slot->identifier == NULL) {
		fprintf(stderr, "INTERNAL ERROR: Name of named slot not initialized.\n");
		exit(RC_EINTERNAL);
	}
	fprintf(yyout, "%sSlot = ", rc_slot->identifier);
	hs_gen_slot(rc_slot, 0);
	fprintf(yyout, "\n");
}

void hs_foreach_gen_named_slot(gpointer data, gpointer user_data) {
	hs_gen_named_slot((rc_slot_t*) data);
}

void hs_gen_unary_operation(int unary_operation) {
	switch(unary_operation) {
		case RC_COUNT:
			fprintf(yyout, "Count ");
			break;
		case RC_SUM:
			fprintf(yyout, "Sum ");
			break;
		case RC_PRODUCT:
			fprintf(yyout, "Product ");
			break;
		case RC_MINIMUM:
			fprintf(yyout, "Minimum ");
			break;
		case RC_MAXIMUM:
			fprintf(yyout, "Maximum ");
			break;
		case RC_SQUARE_ROOT:
			fprintf(yyout, "SquareRoot ");
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid unary arithmetic operation %d.\n", unary_operation);
			exit(RC_EINTERNAL);
	}
}

void hs_gen_binary_operation(int binary_operation) {
	switch(binary_operation) {
		case RC_ADD:
			fprintf(yyout, " Add ");
			break;
		case RC_SUBTRACT:
			fprintf(yyout, " Subtract ");
			break;
		case RC_MULTIPLY:
			fprintf(yyout, " Multiply ");
			break;
		case RC_DIVIDE:
			fprintf(yyout, " Divide ");
			break;
		case RC_TO_THE_POWER_OF:
			fprintf(yyout, " ToThePowerOf ");
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid binary arithmetic operation %d.\n", binary_operation);
			exit(RC_EINTERNAL);
	}
}

void hs_gen_expression(rc_expression_t* rc_expression, iterator_t* iter) {
	switch(rc_expression->type) {
		case RC_CONSTANT:
			fprintf(yyout, "(Constant ");
			hs_gen_variable(rc_expression->constant_variable);
			fprintf(yyout, ")");
			break;
		case RC_UNARY_OPERATION:
			fprintf(yyout, "(UnaryOperation ");
			hs_gen_unary_operation(rc_expression->unary_operation);
			hs_gen_expression(rc_expression->unary_expression, NULL);
			fprintf(yyout, ")");
			break;
		case RC_BINARY_OPERATION:
			fprintf(yyout, "(BinaryOperation ");
			hs_gen_expression(rc_expression->left_binary_expression, NULL);
			hs_gen_binary_operation(rc_expression->binary_operation);
			hs_gen_expression(rc_expression->right_binary_expression, NULL);
			fprintf(yyout, ")");
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid expression type %d.\n", rc_expression->type);
			exit(RC_EINTERNAL);
	}
	if((iter != NULL) && (++iter->i < iter->len))
		fprintf(yyout, ", ");
}

void hs_foreach_gen_expression(gpointer data, gpointer user_data) {
	hs_gen_expression((rc_expression_t*) data, (iterator_t*) user_data);
}

void hs_gen_comparison_operation(int comparison_operation) {
	switch(comparison_operation) {
		case RC_EQUALS:
			fprintf(yyout, " Equals ");
			break;
		case RC_NOT_EQUALS:
			fprintf(yyout, " NotEquals ");
			break;
		case RC_LESS_THAN:
			fprintf(yyout, " LessThan ");
			break;
		case RC_GREATER_THAN:
			fprintf(yyout, " GreaterThan ");
			break;
		case RC_LESS_OR_EQUAL:
			fprintf(yyout, " LessOrEqual ");
			break;
		case RC_GREATER_OR_EQUAL:
			fprintf(yyout, " GreaterOrEqual ");
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid comparison operation type %d.\n", comparison_operation);
			exit(RC_EINTERNAL);
	}
}

void hs_gen_condition(rc_condition_t* rc_condition, iterator_t* iter) {
	switch(rc_condition->type) {
		case RC_EXISTS:
			fprintf(yyout, "(Exists ");
			hs_gen_slot(rc_condition->exists_slot, 1);
			fprintf(yyout, ")");
			break;
		case RC_EVAL:
			fprintf(yyout, "(Eval ");
			hs_gen_expression(rc_condition->eval_left_expression, NULL);
			hs_gen_comparison_operation(rc_condition->eval_comparison_operation);
			hs_gen_expression(rc_condition->eval_right_expression, NULL);
			fprintf(yyout, ")");
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid condition type %d.\n", rc_condition->type);
			exit(RC_EINTERNAL);
	}
	if((iter != NULL) && (++iter->i < iter->len))
		fprintf(yyout, ", ");
}

void hs_foreach_gen_condition(gpointer data, gpointer user_data) {
	hs_gen_condition((rc_condition_t*) data, (iterator_t*) user_data);
}

void hs_gen_initializer(rc_initializer_t* rc_initializer, iterator_t* iter) {
	fprintf(yyout, "(\"%s\", ", rc_initializer->key);
	hs_gen_expression(rc_initializer->expression, NULL);
	fprintf(yyout, ")");
	if((iter != NULL) && (++iter->i < iter->len))
		fprintf(yyout, ", ");
}

void hs_foreach_gen_initializer(gpointer data, gpointer user_data) {
	hs_gen_initializer((rc_initializer_t*) data, (iterator_t*) user_data);
}

void hs_gen_statement(rc_statement_t* rc_statement, iterator_t* iter) {
	switch(rc_statement->type) {
		case RC_DEFINE:
			{
				iterator_t* iter = (iterator_t*) malloc_zero(sizeof(iterator_t));
				iter->i = 0;
				iter->len = g_list_length(rc_statement->define_initializers);
				fprintf(yyout, "(Define (newName namespace \"%s\") [", rc_statement->define_name->string);
				g_list_foreach(rc_statement->define_initializers, hs_foreach_gen_initializer, iter);
				fprintf(yyout, "])");
				free(iter);
			}
			break;
		case RC_RETRACT:
			fprintf(yyout, "(Retract ");
			hs_gen_slot(rc_statement->retract_slot, 1);
			fprintf(yyout, ")");
			break;
		case RC_COPY:
			fprintf(yyout, "(Copy ");
			hs_gen_slot(rc_statement->copy_slot, 1);
			fprintf(yyout, ")");
			break;
		case RC_SEND:
			fprintf(yyout, "(Send ");
			hs_gen_expression(rc_statement->send_address_expression, NULL);
			fprintf(yyout, " ");
			hs_gen_expression(rc_statement->send_power_expression, NULL);
			fprintf(yyout, " ");
			hs_gen_slot(rc_statement->send_slot, 1);
			fprintf(yyout, ")");
			break;
		case RC_SET:
			fprintf(yyout, "(Set ");
			hs_gen_slot(rc_statement->set_slot, 1);
			fprintf(yyout, " ");
			hs_gen_expression(rc_statement->set_expression, NULL);
			fprintf(yyout, ")");
			break;
		case RC_FLUSH:
			fprintf(yyout, "(Flush ");
			hs_gen_slot(rc_statement->flush_slot, 1);
			fprintf(yyout, ")");
			break;
		case RC_CALL:
			{
				iterator_t* iter = (iterator_t*) malloc_zero(sizeof(iterator_t));
				iter->i = 0;
				iter->len = g_list_length(rc_statement->call_expressions);
				fprintf(yyout, "(Call \"%s\" [", rc_statement->call_identifier);
				g_list_foreach(rc_statement->call_expressions, hs_foreach_gen_expression, iter);
				fprintf(yyout, "])");
				free(iter);
			}
			break;
		default:
			fprintf(stderr, "INTERNAL ERROR: Invalid statement type %d.\n", rc_statement->type);
			exit(RC_EINTERNAL);
	}
	if((iter != NULL) && (++iter->i < iter->len))
		fprintf(yyout, ", ");
}

void hs_foreach_gen_statement(gpointer data, gpointer user_data) {
	hs_gen_statement((rc_statement_t*) data, (iterator_t*) user_data);
}

void hs_gen_rule(rc_rule_t* rc_rule) {
	iterator_t* iter = (iterator_t*) malloc_zero(sizeof(iterator_t));

	fprintf(yyout, "%sRule = newRule \"%s\" %d [", rc_rule->identifier, rc_rule->identifier, rc_rule->priority);

	iter->i = 0;
	iter->len = g_list_length(rc_rule->conditions);
	g_list_foreach(rc_rule->conditions, hs_foreach_gen_condition, iter);
	fprintf(yyout, "] [");

	iter->i = 0;
	iter->len = g_list_length(rc_rule->statements);
	g_list_foreach(rc_rule->statements, hs_foreach_gen_statement, iter);
	fprintf(yyout, "]\n\n");

	free(iter);
}

void hs_foreach_gen_rule(gpointer data, gpointer user_data) {
	hs_gen_rule((rc_rule_t*) data);
}

void hs_gen_property(rc_property_t* rc_property, iterator_t* iter) {
	fprintf(yyout, "(newProperty \"%s\" ", rc_property->key);
	hs_gen_variable(rc_property->variable);
	fprintf(yyout, ")");
	if((iter != NULL) && (++iter->i < iter->len))
		fprintf(yyout, ", ");
}

void hs_foreach_gen_property(gpointer data, gpointer user_data) {
	hs_gen_property((rc_property_t*) data, (iterator_t*) user_data);
}

void hs_gen_fact(rc_fact_t* rc_fact) {
	iterator_t* iter = (iterator_t*) malloc_zero(sizeof(iterator_t));

	fprintf(yyout, "%sFact_%d = newFact \"%s\" [", rc_fact->name->string, rc_fact->unique_id, rc_fact->name->string);

	iter->i = 0;
	iter->len = g_list_length(rc_fact->properties);
	g_list_foreach(rc_fact->properties, hs_foreach_gen_property, iter);
	fprintf(yyout, "]\n");

	free(iter);
}

void hs_foreach_gen_fact(gpointer data, gpointer user_data) {
	hs_gen_fact((rc_fact_t*) data);
}

void hs_foreach_gen_depends(gpointer data, gpointer user_data) {
	fprintf(yyout, "import qualified Rulesets.%s as %s\n", (char*) data, (char*) data);
}

void hs_gen_ruleset(rc_ruleset_t* rc_ruleset) {
	cur_ruleset = rc_ruleset;

	fprintf(yyout, "--\n");
	fprintf(yyout, "-- This file was generated automatically from \"%s\" by facts-rc.\n", in_filename);
	fprintf(yyout, "-- It should be located in the Rulesets/ directory of your FACTS-hs installation.\n");
	fprintf(yyout, "--\n");
	fprintf(yyout, "\n");
	fprintf(yyout, "module Rulesets.%s (\n", rc_ruleset->identifier);
	fprintf(yyout, "\n");
	fprintf(yyout, "	ruleset");
	{
		GList* elem = g_list_first(rc_ruleset->names);
		while(elem != NULL) {
			rc_name_t* name = (rc_name_t*) elem->data;
			fprintf(yyout, ", %sName", name->identifier);
			elem = g_list_next(elem);
		}
	}
	{
		GList* elem = g_list_first(rc_ruleset->slots);
		while(elem != NULL) {
			rc_slot_t* slot = (rc_slot_t*) elem->data;
 			fprintf(yyout, ", %sSlot", slot->identifier);
			elem = g_list_next(elem);
		}
	}
	fprintf(yyout, "\n");
	fprintf(yyout, "\n");
	fprintf(yyout, ") where\n");
	fprintf(yyout, "\n");
	fprintf(yyout, "import Facts\n");
	g_list_foreach(rc_ruleset->depends, hs_foreach_gen_depends, NULL);
	fprintf(yyout, "\n");
	fprintf(yyout, "namespace = \"%s\"\n", rc_ruleset->identifier);
	fprintf(yyout, "\n");
	fprintf(yyout, "ruleset = newRuleset namespace [");
	{
		GList* elem = g_list_first(rc_ruleset->rules);
		while(elem != NULL) {
			rc_rule_t* rc_rule = (rc_rule_t*) elem->data;
			fprintf(yyout, "%sRule", rc_rule->identifier);
			elem = g_list_next(elem);
			if(elem != NULL)
				fprintf(yyout, ", ");
		}
	}
	fprintf(yyout, "] [");
	{
		GList* elem = g_list_first(rc_ruleset->facts);
		while(elem != NULL) {
			rc_fact_t* rc_fact = (rc_fact_t*) elem->data;
			fprintf(yyout, "%sFact_%d", rc_fact->name->string, rc_fact->unique_id);
			elem = g_list_next(elem);
			if(elem != NULL)
				fprintf(yyout, ", ");
		}
	}
	fprintf(yyout, "] [");
	{
		GList* elem = g_list_first(rc_ruleset->functions);
		while(elem != NULL) {
			rc_function_t* rc_function = (rc_function_t*) elem->data;
			fprintf(yyout, "%sFunction", rc_function->identifier);
			elem = g_list_next(elem);
			if(elem != NULL)
				fprintf(yyout, ", ");
		}
	}
	fprintf(yyout, "]\n");
	fprintf(yyout, "\n");
	g_list_foreach(rc_ruleset->names, hs_foreach_gen_named_name, NULL);
	fprintf(yyout, "\n");
	g_list_foreach(rc_ruleset->slots, hs_foreach_gen_named_slot, NULL);
	fprintf(yyout, "\n");
	g_list_foreach(rc_ruleset->facts, hs_foreach_gen_fact, NULL);
	fprintf(yyout, "\n");
	g_list_foreach(rc_ruleset->rules, hs_foreach_gen_rule, NULL);
	fprintf(yyout, "\n");
}

void hs_foreach_gen_ruleset(gpointer data, gpointer user_data) {
	hs_gen_ruleset((rc_ruleset_t*) data);
}

void hs_generate(GList* rulesets) {
	g_list_foreach(rulesets, hs_foreach_gen_ruleset, NULL);
}
