/*
 * Copyright 2005, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * Contributors: Georg Wittenburg
 */

#include "main.h"
#include "free.h"
#include "hs_gen.h"
#include "eeprom_gen.h"
#include "y.tab.h"

/**
 * This is the main file of the rule compiler.
 **/

void usage() {
	fprintf(stderr, "Usage: facts-rc [-v] [-d] [-e file|-h] [-O] [-o file] file\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -v\t\tPrint version and exit.\n");
	fprintf(stderr, "  -d\t\tPrint debugging information.\n");
	fprintf(stderr, "  -e file\tInput EEPROM image file.\n");
	fprintf(stderr, "  -h\t\tCompile Haskell code for FACTS-hs rather than bytecode.\n");
	fprintf(stderr, "  -O\t\tOptimize size of bytecode.\n");
	fprintf(stderr, "  -o file\tOutput file (either EEPROM bytecode or Haskell).\n");
	fprintf(stderr, "  file\t\tInput ruleset file.\n");
	exit(RC_EPARAM);
}

int main(int argc, char *argv[]) {
	int output_to_eeprom = 1;
	char* out_filename = NULL;
	char* eeprom_filename = NULL;
	FILE* eeprom_file = NULL;

	// Initialize global variables.
	in_filename = NULL;
	line = 1;
	errors = 0;
	optimize_bytecode = 0;
	rulesets = NULL;
	debug = 0;

	int c;
	opterr = 0;
	while((c = getopt(argc, argv, "de:ho:Ov")) != -1) {
		switch(c) {
			case 'd':
				debug = 1;
				break;
			case 'e':
				eeprom_filename = optarg;
				break;
			case 'h':
				output_to_eeprom = 0;
				break;
			case 'o':
				out_filename = optarg;
				break;
			case 'O':
				optimize_bytecode = 1;
				break;
			case 'v':
				fprintf(stderr, "facts-rc, the FACTS Ruleset Compiler, version %s.\n", RC_VERSION);
				exit(0);
			case '?':
				if(isprint(optopt))
					fprintf(stderr, "Unknown option `-%c'.\n", optopt);
				else
					fprintf(stderr, "Unknown option character `\\x%x'.\n", optopt);
				return RC_EPARAM;
			default:
				abort();
		}
	}
	if(optind < argc)
		in_filename = argv[optind];

	if(
		(output_to_eeprom && (eeprom_filename == NULL))
		|| (in_filename == NULL)
		|| ((optind + 1) < argc)
	) usage();

	if(out_filename == NULL) {
		if(output_to_eeprom)
			out_filename = "out.bin";
		else
			out_filename = "out.hs";
	}

	// Open files.
	if((yyin = fopen(in_filename, "r")) == NULL) {
		fprintf(stderr, "Unable to open file \"%s\" for reading.\n", in_filename);
		exit(RC_EIO);
	}
	if((yyout = fopen(out_filename, "w")) == NULL) {
		fprintf(stderr, "Unable to open file \"%s\" for writing.\n", out_filename);
		exit(RC_EIO);
	}
	if(output_to_eeprom && ((eeprom_file = fopen(eeprom_filename, "r")) == NULL)) {
		fprintf(stderr, "Unable to open file \"%s\" for reading.\n", eeprom_filename);
		exit(RC_EIO);
	}

	// Parse the input file, constructing the data tree in rulesets.
	yyparse();

	// Generate output or abort on errors.
	if(errors == 0) {
		if(debug)
			fprintf(stderr, "Parsed %d lines.\n\n", line);
		if(output_to_eeprom)
			eeprom_generate(rulesets, eeprom_file);
		else
			hs_generate(rulesets);
	} else {
		fprintf(stderr, "Parsed %d lines with %d errors.\n", line, errors);
	}

	// Close files.
	if(
		(fclose(yyin) != 0)
		|| (fclose(yyout) != 0)
		|| (output_to_eeprom && (fclose(eeprom_file) != 0))
	) {
		fprintf(stderr, "Unable to close file.\n");
		exit(RC_EIO);
	}

	// Free memory only if there were no errors while parsing or if we're
	// running debug mode.
	if((errors == 0) || debug) {
		g_list_foreach(rulesets, foreach_free_ruleset, NULL);
		g_list_free(rulesets);
	}

	if(errors != 0)
		exit(RC_ECOMPILE);
	return 0;
}

void* malloc_zero(size_t size) {
	void* mem = malloc(size);
	if(mem == NULL) {
		fprintf(stderr, "Failed to allocate enough memory.\n");
		exit(RC_ENOMEM);
	}
	memset(mem, 0, size);
	return mem;
}
