/*
 * Copyright 2005, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * Contributors: Georg Wittenburg
 */

#ifndef MAIN_H
#define MAIN_H 1

#include <ctype.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// We need to include this header for EEPROM related typedefs.
#include "eeprom_gen.h"



#define RC_VERSION	"0.3"

#define RC_EPARAM	1	// Exit due to invalid parameters.
#define RC_ECOMPILE	2	// Exit due to compile errors.
#define RC_EEEPROM	3	// Exit due to invalid EEPROM layout.
#define RC_EIO		4	// Exit due to I/O errors.
#define RC_ENOMEM	5	// Exit due to unable to allocate memory.
#define RC_EINTERNAL	6	// Exit due to internal errors. Shouldn't happen.



// Make some Flex / Bison functions visible in the entire project.
extern int yylex();
extern int yyparse();
extern void yyerror(char* msg);
extern FILE *yyin, *yyout;
extern char* yytext;



/**
 * The following structs are the central means of data storage of the compiler.
 * They are created while parsing the input ruleset and used by the backends to
 * generate their respective output.
 **/
 
typedef struct {
	char* identifier;
	char* string;
} rc_name_t;

typedef struct {
	int type;
	void* value;
} rc_variable_t;

typedef struct {
	char* identifier;
	rc_name_t* name;
	char* key;
	GList* conditions;
} rc_slot_t;

// Defines for rc_expression_t.type.
#define RC_CONSTANT		0
#define RC_UNARY_OPERATION	1
#define RC_BINARY_OPERATION	2
typedef struct _rc_expression_t {
	int type;
	// RC_CONSTANT fields.
	rc_variable_t* constant_variable;
	// RC_UNARY_OPERATION fields.
	int unary_operation;
	struct _rc_expression_t* unary_expression;
	// RC_BINARY_OPERATION fields.
	int binary_operation;
	struct _rc_expression_t* left_binary_expression;
	struct _rc_expression_t* right_binary_expression;
} rc_expression_t;

typedef struct {
	int type;
	// EXISTS fields.
	rc_slot_t* exists_slot;
	// EVAL fields.
	rc_expression_t* eval_left_expression;
	int eval_comparison_operation;
	rc_expression_t* eval_right_expression;
} rc_condition_t;

typedef struct {
	char* key;
	rc_expression_t* expression;
} rc_initializer_t;

typedef struct {
	int type;
	// DEFINE fields.
	rc_name_t* define_name;
	GList* define_initializers;
	// RETRACT fields.
	rc_slot_t* retract_slot;
	// COPY fields.
	rc_slot_t* copy_slot;
	// SEND fields.
	rc_expression_t* send_address_expression;
	rc_expression_t* send_power_expression;
	rc_slot_t* send_slot;
	// SET fields.
	rc_slot_t* set_slot;
	rc_expression_t* set_expression;
	// FLUSH fields.
	rc_slot_t* flush_slot;
	// CALL fields.
	char* call_identifier;
	GList* call_expressions;
} rc_statement_t;

typedef struct {
	char* identifier;
	int priority;
	GList* conditions;
	GList* statements;
} rc_rule_t;

typedef struct {
	char* key;
	rc_variable_t* variable;
} rc_property_t;

typedef struct {
	rc_name_t* name;
	GList* properties;
	int unique_id;
} rc_fact_t;

typedef struct {
	char* identifier;
} rc_function_t;

typedef struct {
	char* identifier;
	GList* depends;
	GList* facts;
	int facts_count;
	GList* rules;
	int rules_count;
	GList* functions;
	int functions_count;
	GList* slots;
	int slots_count;
	GList* names;
	int names_count;
	int expressions_count;
	int conditions_count;
	int statements_count;
	int initializers_count;
	int const_variables_count;
	int properties_count;
} rc_ruleset_t;

// This is a special constant used to model the "this" keyword referring to the
// current fact in the process of slot evaluation.
static rc_name_t rc_name_this = {
	"this",
	"this"
};

void* malloc_zero(size_t size);	// Utility function to allocate and zero memory.
char* in_filename;		// The file name of the input file.
int line;			// The line that we're parsing right now.
int errors;			// Amount of errors encountered while parsing.
int optimize_bytecode;		// Optimize the output.
GList* rulesets;		// The root of the parsed data tree.
int debug;			// Display debugging output.

#endif // MAIN_H
